from requests import post
from fastapi import HTTPException

from ..constants import PAYPAL_API_URL, PAYPAL_CLIENT_ID, PAYPAL_SECRET


def get_paypal_access_token():
    """
    Get an access token from PayPal to authenticate API requests.
    """
    response = post(
        f"{PAYPAL_API_URL}/v1/oauth2/token",
        headers={
            "Accept": "application/json",
            "Accept-Language": "en_US",
        },
        data={"grant_type": "client_credentials"},
        auth=(PAYPAL_CLIENT_ID, PAYPAL_SECRET),
    )

    if response.status_code != 200:
        raise HTTPException(status_code=response.status_code, detail="Failed to get PayPal access token")
    
    return response.json()["access_token"]


def create_paypal_order(amount: float, currency: str = "USD"):
    """
    Create a PayPal payment order.
    """
    access_token = get_paypal_access_token()

    order_data = {
        "intent": "CAPTURE",
        "purchase_units": [
            {
                "amount": {
                    "currency_code": currency,
                    "value": f"{amount:.2f}",
                }
            }
        ],
        "application_context": {
            "return_url": "https://placeholderlink.com/payment-success",
            "cancel_url": "https://placeholderlink.com/payment-cancelled",
        }
    }

    response = post(
        f"{PAYPAL_API_URL}/v2/checkout/orders",
        headers={
            "Content-Type": "application/json",
            "Authorization": f"Bearer {access_token}",
        },
        json=order_data,
    )

    if response.status_code != 201:
        raise HTTPException(status_code=response.status_code, detail="Failed to create PayPal order")

    order = response.json()
    approval_url = next(link['href'] for link in order['links'] if link['rel'] == 'approve')

    return {
        "approval_url": approval_url,
        "order_id": order["id"],
    }


def capture_paypal_order(order_id: str):
    """
    Capture the PayPal payment order after user approval.
    """
    access_token = get_paypal_access_token()

    response = post(
        f"{PAYPAL_API_URL}/v2/checkout/orders/{order_id}/capture",
        headers={
            "Content-Type": "application/json",
            "Authorization": f"Bearer {access_token}",
        },
    )

    if response.status_code != 201:
        raise HTTPException(status_code=response.status_code, detail="Failed to capture PayPal payment")

    return response.json()
