from fastapi import HTTPException, status
from supabase import create_client, Client
from ..constants import SUPABASE_URL, SUPABASE_KEY

class Supabase:
    def __init__(self):
        # Initialize Supabase client
        print(SUPABASE_URL)
        self.supabase: Client = create_client(SUPABASE_URL, SUPABASE_KEY)


    def check_balance(self, user_id: str):
        """
        Checks if the user has enough 'credit' for generating a new trip.

        Args:
            user_id (str): The ID of the current user.

        Returns:
            bool: True if the balance is positive, False otherwise.
        """
        try:
            # Query the public.credits table to get the user's credits
            response = self.supabase.table('credits').select('credits').eq('user_id', user_id).single().execute()

            # Check if the response was successful and contains data
            if response.get('data') and 'credits' in response['data']:
                credits = response['data']['credits']
                
                if credits <= 0: 
                    raise HTTPException(status_code=403, detail="Insufficient credits to perform this action.")


            # If user or credits not found, raise a 404 error
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="User data not found.")

        except Exception:
            # Handle exceptions (like if the user ID doesn't exist or database issues)
            raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Error retrieving credit balance")


    def decrement_credit_balance(self, user_id: str) -> bool:
        """
        Decrements the user's credit balance by 1.

        Args:
            user_id (str): The ID of the current user.
        """
        try:
            # Query the public.credits table to get the user's current credits
            response = self.supabase.table('credits').select('credits').eq('user_id', user_id).single().execute()

            # Check if the response was successful and contains data
            if response.get('data') and 'credits' in response['data']:
                current_credits = response['data']['credits']
                # Decrement the credit balance by 1
                updated_credits = current_credits - 1

                # Update the user's credits in the table
                update_response = self.supabase.table('credits').update({'credits': updated_credits}).eq('user_id', user_id).execute()

                # Check if the update was successful
                if update_response.get('status_code') == 200:
                    return True
                else:
                    raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Error updating credit balance.")

            # If user or credits not found, raise a 404 error
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="User data not found.")

        except Exception:
            # Handle exceptions (like if the user ID doesn't exist or there are database issues)
            raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Error updating credit balance")


    def increment_credit_balance(self, user_id: str) -> bool:
        """
        Increments the user's credit balance by 1.

        Args:
            user_id (str): The ID of the current user.
        """
        try:
            # Query the public.credits table to get the user's current credits
            response = self.supabase.table('credits').select('credits').eq('user_id', user_id).single().execute()

            # Check if the response was successful and contains data
            if response.get('data') and 'credits' in response['data']:
                current_credits = response['data']['credits']
                # Increment the credit balance by 1
                updated_credits = current_credits + 1

                # Update the user's credits in the table
                update_response = self.supabase.table('credits').update({'credits': updated_credits}).eq('user_id', user_id).execute()

                # Check if the update was successful
                if update_response.get('status_code') == 200:
                    return True
                else:
                    raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Error updating credit balance.")

            # If user or credits not found, raise a 404 error
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="User data not found.")

        except Exception:
            # Handle exceptions (like if the user ID doesn't exist or there are database issues)
            raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Error updating credit balance")
